
/*************** Definitions ******************/

abstract sig Entity{}

sig User extends Entity{
	pClient : one PrefetchingClient,
	viewableMessages : set Message,
	headlessBrowser : one HeadlessBrowser
}
{
	all m : Message { m.u = this iff m in viewableMessages }
}

sig AttackerEavesdropper extends User{
	attackerData : set Data
}

sig PrefetchingClient extends Entity{	
	user : one User,
	cache : one Cache
}
{
	cache not in PrefetchingProxy.sharedCache
}

one sig PrefetchingProxy extends Entity{
	sharedCache : one Cache,
	headlessBrowser : set HeadlessBrowser
}
{
	all hb : HeadlessBrowser { hb in headlessBrowser }
	sharedCache not in PrefetchingClient.cache
	sharedCache.data = HTMLResponse.content + embeddedObject.content
}

sig HeadlessBrowser extends Entity{
	hbCache : one Cache
}
{
	hbCache = PrefetchingProxy.sharedCache
}

sig Link extends Entity{}
one sig SatelliteLink, InternetLink extends Link{}

sig WebServer extends Entity{}

sig AttackerWebServer extends WebServer{
	malware : Data
}

sig Cache {
	data : set Data
}

sig Data{}

abstract sig Message{
	to : Entity,
	from : Entity,
	content : Data,
	u : one User,
	pc : one PrefetchingClient,
	w : one WebServer
}
{
	from != to
	u.pClient = pc
	pc.user = u
	to in User => { to = u }
	from in User => { from = u }
	to in PrefetchingClient => { to = pc }
	from in PrefetchingClient => { from = pc }
	to in WebServer => { to = w }
	from in WebServer => { from = w }
}

sig embeddedObject extends Message{}
{
	from not in PrefetchingClient + SatelliteLink + PrefetchingProxy + 
		PrefetchingProxy.headlessBrowser + InternetLink + WebServer
		iff to = none
	from in PrefetchingClient iff to = u
	from in SatelliteLink iff to = pc
	from in PrefetchingProxy iff to = SatelliteLink
	from in PrefetchingProxy.headlessBrowser iff to = PrefetchingProxy
	from in InternetLink iff to in PrefetchingProxy.headlessBrowser
	from in WebServer iff to = InternetLink

	w not in AttackerWebServer => {
		all a : AttackerWebServer {
			content != a.malware
		}
	} else {
		some a : AttackerWebServer {
			content = a.malware
		}
	}

	content in (pc.cache).data
}

sig GETembedded extends Message{}
{
	from not in User + PrefetchingProxy.headlessBrowser + InternetLink iff to = none
	from in User iff to = pc
	from in PrefetchingProxy.headlessBrowser iff to = InternetLink
	from in InternetLink iff to = w

	content in (pc.cache).data
}

sig GET extends Message{}
{
	from in User iff to = pc
	from in PrefetchingClient iff  to = SatelliteLink
	from in SatelliteLink iff to = PrefetchingProxy
	from in PrefetchingProxy iff to in PrefetchingProxy.headlessBrowser
	from in PrefetchingProxy.headlessBrowser iff to = InternetLink
	from in InternetLink iff to = w
	from in WebServer iff to = none

	content in (pc.cache).data
}

sig HTMLResponse extends Message{}
{
	from in WebServer iff to = InternetLink
	from in InternetLink iff to = PrefetchingProxy
	from in PrefetchingProxy iff to in PrefetchingProxy.headlessBrowser
	from in PrefetchingProxy.headlessBrowser iff to = SatelliteLink
	from in SatelliteLink iff to = pc
	from in PrefetchingClient iff to = u
	from in User iff to = none

	w not in AttackerWebServer => {
		all a : AttackerWebServer {
			content != a.malware
		}
	} else {
		some a : AttackerWebServer {
			content = a.malware
		}
	}

	content in (pc.cache).data
}

/****************** Facts / Predicates ******************/

fact { 
	all u1, u2 : User {
		u1 != u2 => {
			u1.pClient != u2.pClient
			u1.headlessBrowser != u2.headlessBrowser
		}
	}
}

fact { 
	all g1, g2 : GET {
		g1 != g2 and
		g1.u = g2.u => {
			g1.content != g2.content
			g1.pc = g2.pc
			some disj h1 , h2 : HTMLResponse {
				g1.u = h1.u
				g1.u = h2.u
				h1.content != h2.content
				h1.pc = h2.pc
				h1.pc = g1.pc
			}
			some disj ge1, ge2 : GETembedded {
				g1.u = ge1.u
				g1.u = ge2.u
				ge1.content != ge2.content
				ge1.pc = ge2.pc
				ge1.pc = g1.pc
			}
			some disj eo1, eo2 : embeddedObject {
				g1.u = eo1.u
				g1.u = eo2.u
				eo1.content != eo2.content
				eo1.pc = eo2.pc
				eo1.pc = g1.pc
			}
		}
	}
}

fact { 
	all pc : PrefetchingClient, u : User {
		pc = u.pClient iff
			pc.user = u
	}
}

fact { 
	all m1, m2 : Message {
		m1 != m2 => {
			m1 in GET and m2 not in GET => {
				m1.content != m2.content
			}
			m1 in HTMLResponse and m2 not in HTMLResponse => {
				m1.content != m2.content
			}
			m1 in GETembedded and m2 not in GETembedded => {
				m1.content != m2.content
			}
			m1 in embeddedObject and m2 not in embeddedObject => {
				m1.content != m2.content
			}
		}
	}
}

fact { 
	all g1, g2 : GET {
		g1 != g2 and
		g1.content != g2.content => {
			g1.w != g2.w
			some disj h1, h2 : HTMLResponse {
				g1.u = g2.u => {
					g1.pc = g2.pc
				} else {
					g1.pc != g2.pc
				}
				h1.pc = g1.pc
				h2.pc = g2.pc
				h1.u = g1.u
				h2.u = g2.u

				h1.content != h2.content
				h1.w != g1.w
				h2.w != g1.w

				g1.u in AttackerEavesdropper and g2.u not in AttackerEavesdropper => {
					g2.content not in (g1.u).attackerData
					h2.content not in (g1.u).attackerData
				}
			}
			some disj ge1, ge2 : GETembedded {
				g1.u = g2.u => {
					g1.pc = g2.pc
				} else {
					g1.pc != g2.pc
				}
				ge1.pc = g1.pc
				ge2.pc = g2.pc
				ge1.u = g1.u
				ge2.u = g2.u

				ge1.content != ge2.content
				ge1.w != g1.w
				ge2.w != g1.w

				g1.u in AttackerEavesdropper and g2.u not in AttackerEavesdropper => {
					g2.content not in (g1.u).attackerData
					ge2.content not in (g1.u).attackerData
				}
			}
			some disj eo1, eo2 : embeddedObject {
				g1.u = g2.u => {
					g1.pc = g2.pc
				} else {
					g1.pc != g2.pc
				}
				eo1.pc = g1.pc
				eo2.pc = g2.pc
				eo1.u = g1.u
				eo2.u = g2.u

				eo1.content != eo2.content
				eo1.w != g1.w
				eo2.w != g1.w

				g1.u in AttackerEavesdropper and g2.u not in AttackerEavesdropper => {
					g2.content not in (g1.u).attackerData
					eo2.content not in (g1.u).attackerData
				}
			}
		}
	}
}

fact { 
	all g1, g2 : GET {
		g1 != g2 and
		g1.content = g2.content => {
			g1.w = g2.w
			some disj h1, h2 : HTMLResponse {
				g1.u = g2.u => {
					g1.pc = g2.pc
				} else {
					g1.pc != g2.pc
				}
				h1.pc = g1.pc
				h2.pc = g2.pc
				h1.u = g1.u
				h2.u = g2.u

				h1.content = h2.content
				h1.w = g1.w
				h2.w = g1.w

				g1.u in AttackerEavesdropper and g2.u not in AttackerEavesdropper => {
					g2.content in (g1.u).attackerData
					h2.content in (((g1.u).headlessBrowser).hbCache).data => {
						h2.content in (g1.u).attackerData
					}
				}
			}
			some disj ge1, ge2 : GETembedded {
				g1.u = g2.u => {
					g1.pc = g2.pc
				} else {
					g1.pc != g2.pc
				}
				ge1.pc = g1.pc
				ge2.pc = g2.pc
				ge1.u = g1.u
				ge2.u = g2.u

				ge1.content = ge2.content
				ge1.w = g1.w
				ge2.w = g1.w

				g1.u in AttackerEavesdropper and g2.u not in AttackerEavesdropper => {
					g2.content in (g1.u).attackerData
					ge2.content in (((g1.u).headlessBrowser).hbCache).data => {
						ge2.content in (g1.u).attackerData
					}
				}
			}
			some disj eo1, eo2 : embeddedObject {
				g1.u = g2.u => {
					g1.pc = g2.pc
				} else {
					g1.pc != g2.pc
				}
				eo1.pc = g1.pc
				eo2.pc = g2.pc
				eo1.u = g1.u
				eo2.u = g2.u

				eo1.content = eo2.content
				eo1.w = g1.w
				eo2.w = g1.w

				g1.u in AttackerEavesdropper and g2.u not in AttackerEavesdropper => {
					g2.content in (g1.u).attackerData
					eo2.content in (((g1.u).headlessBrowser).hbCache).data => {
						eo2.content in (g1.u).attackerData
					}
				}
			}
		}
	}
}

fact {
	some m : Message, d : Data {
		d != m.content => {
			d not in Cache.data
		}
	}
}

fact { 
	all pc1, pc2 : PrefetchingClient {
		pc1 != pc2 => {
			pc1.cache != pc2.cache
			pc1.user != pc2.user
		}
	}
}

assert checkGetsFromDiffUser {
	all g1, g2 : GET {
		g1 != g2 => {
			g1.u != g2.u
			g1.pc != g2.pc
		}
	}
}

assert checkHTMLResponseToDiffUser {
	all h1, h2 : HTMLResponse {
		h1 != h2 => {
			h1.u != h2.u
			h1.pc != h2.pc
		}
	}
}

assert checkGetsEmbeddedToDiffUser {
	all g1, g2 : GETembedded {
		g1 != g2 => {
			g1.u != g2.u
			g1.pc != g2.pc
		}
	}
}

assert checkEmbeddedObjectFromDiffUser {
	all e1, e2 : embeddedObject {
		e1 != e2 => {
			e1.u != e2.u
			e1.pc != e2.pc
		}
	}
}

assert checkPrefetchingClientCacheUnique {
	all pc1, pc2 : PrefetchingClient {
		pc1 != pc2 => {
			(pc1.cache).data & (pc2.cache).data = none
		}
	}
}

assert checkMessageContentsUnique {
	all m1, m2 : Message {
		m1 != m2 => {
			m1.content != m2.content
		}
	}
}

assert checkUserViewableMessages {
	all u1, u2 : User {
		u1 != u2 => {
			u1.viewableMessages & u2.viewableMessages = none
		}
	}
}

assert checkTwoGetsSameRequest {
	all g1, g2 : GET {
		g1 != g2 => {
			g1.content = g2.content => {
				all h1, h2 : HTMLResponse {
					h1 != h2 => {
						(h1 in (g1.u).viewableMessages) && (h2 in (g2.u).viewableMessages) => {
							h1.content = h2.content
						}
					}
				}
				all ge1, ge2 : GETembedded {
					ge1 != ge2 => {
						(ge1 in (g1.u).viewableMessages) && (ge2 in (g2.u).viewableMessages) => {
							ge1.content = ge2.content
						}
					}
				}
				all eo1, eo2 : embeddedObject {
					eo1 != eo2 => {
						(eo1 in (g1.u).viewableMessages) && (eo2 in (g2.u).viewableMessages) => {
							eo1.content = eo2.content
						}
					}
				}
			}
		}
	}
}

check checkGetsFromDiffUser 
for 11 Entity, exactly 2 User, exactly 0 AttackerEavesdropper, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 8 Data, 
exactly 8 Message, 2 GETembedded, 2 embeddedObject, 2 GET, 2 HTMLResponse

check checkHTMLResponseToDiffUser
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 2 Data, 
exactly 2 Message, 0 GETembedded, 0 embeddedObject, 0 GET, 2 HTMLResponse

check checkGetsEmbeddedToDiffUser
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 2 Data, 
exactly 2 Message, 2 GETembedded, 0 embeddedObject, 0 GET, 0 HTMLResponse

check checkEmbeddedObjectFromDiffUser
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 2 Data, 
exactly 2 Message, 0 GETembedded, 2 embeddedObject, 0 GET, 0 HTMLResponse

check checkPrefetchingClientCacheUnique
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 2 Data, 
exactly 2 Message, 0 GETembedded, 0 embeddedObject, 0 GET, 2 HTMLResponse

check checkMessageContentsUnique
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 2 Data, 
exactly 2 Message, 0 GETembedded, 0 embeddedObject, 0 GET, 2 HTMLResponse


check checkUserViewableMessages
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 4 Data, 
exactly 4 Message, 0 GETembedded, 0 embeddedObject, 2 GET, 2 HTMLResponse

check checkTwoGetsSameRequest
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 8 Data, 
exactly 8 Message, 2 GETembedded, 2 embeddedObject, 2 GET, 2 HTMLResponse

pred sameRequest {
	all g1, g2 : GET {
		g1 != g2 => {
			g1.content = g2.content
		}
	}
}

run sameRequest
for 11 Entity, exactly 2 User, exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 4 Data, 
exactly 4 Message, 0 GETembedded, 0 embeddedObject, 2 GET, 2 HTMLResponse

pred eavesdroppingAttack {
	one attacker : AttackerEavesdropper, user : User {
		one g1, g2 : GET {
			g1.content = g2.content and
			g1.u = attacker and
			g2.u = user and
			g1.w = g2.w => {
				attacker.attackerData != none
				attacker.attackerData = ((user.pClient).cache).data
			} else {
				attacker.attackerData = none
			}
		}
	}
}

run eavesdroppingAttack
for 11 Entity, exactly 2 User, exactly 1 AttackerEavesdropper, 
exactly 2 PrefetchingClient, exactly 2 HeadlessBrowser, exactly 2 WebServer,
exactly 3 Cache, 8 Data, 
exactly 8 Message, 2 GETembedded, 2 embeddedObject, 2 GET, 2 HTMLResponse

pred malwareAttack {
	some attackerSite : AttackerWebServer, user : User {
		some g1 : GET {
			g1.w = attackerSite and
			g1.u = user => {
				attackerSite.malware in ((user.pClient).cache).data
			}
		}
	}
}

run malwareAttack
for 7 Entity, exactly 1 User, exactly 0 AttackerEavesdropper, 
exactly 1 PrefetchingClient, exactly 1 HeadlessBrowser, exactly 1 WebServer, 
exactly 1 AttackerWebServer,  
exactly 2 Cache, 4 Data, 
exactly 4 Message, 1 GETembedded, 1 embeddedObject, 1 GET, 1 HTMLResponse
